#include "pichart.h"

#define N 32768
#define F 8
#define T 10
#define K 16384
static double dt, expmdt, forcdt;
static double *restrict xb,*restrict xgoldb,*restrict yb;
static double *restrict x,*restrict xgold,*restrict y;
static int ncpu;

static void lorenzserial(double *z){
    for(int k=0;k<K;k++){
        z[-1]=z[N-1]; z[-2]=z[N-2]; z[N]=z[0];
        for(int i=0;i<N;i++){
            y[i]=(z[i]+dt*(z[i+1]-z[i-2])*z[i-1])*expmdt+forcdt;
        }
        double *temp=y; y=z; z=temp;
    }
}
static void lorenzblock(double *z,double *w[2],int imin,int imax,int step){
    if(step==1){
        for(int i=imin;i<imax;i++){
            y[i]=(z[i]+dt*(z[i+1]-z[i-2])*z[i-1])*expmdt+forcdt;
        }
    } else {
        double *t0=w[0],*t1=w[1];
        double *zo=z+imin,*yo=y+imin;
        int ilen=imax-imin; 
        int d=step-1;
        for(int i=-2*d;i<ilen+d;i++){
            t0[i]=(zo[i]+dt*(zo[i+1]-zo[i-2])*zo[i-1])*expmdt+forcdt;
        }
        for(--d;d>0;d--){
            for(int i=-2*d;i<ilen+d;i++){
                t1[i]=(t0[i]+dt*(t0[i+1]-t0[i-2])*t0[i-1])*expmdt+forcdt;
            }
            double *temp=t1; t1=t0; t0=temp;
        }
        for(int i=0;i<ilen;i++){
            yo[i]=(t0[i]+dt*(t0[i+1]-t0[i-2])*t0[i-1])*expmdt+forcdt;
        }
    }
}
static void lorenzparallel(double *z){
    double *wb[ncpu][2];
    int w=(N/ncpu+1+3*ncpu)*sizeof(double);
    for(int c=0;c<ncpu;c++){
        wb[c][0]=(double *)mymalloc(w,"wb[c][0]")+2*ncpu;
        wb[c][1]=(double *)mymalloc(w,"wb[c][1]")+2*ncpu;
    }
    double *yorig=y;
    int step=ncpu;
    for(int k=0;k<K;k+=step){
        if(k+step>K) step=K-k;
        for(int j=0;j<step;j++){
            z[N+j]=z[j]; z[-2*j-1]=z[N-2*j-1]; z[-2*j-2]=z[N-2*j-2];
        }
        for(int c=0;c<ncpu;c++){
            int imin=c*N/ncpu,imax=(c+1)*N/ncpu;
            spawn(,firstprivate(c,imin,imax)) 
                lorenzblock(z,wb[c],imin,imax,step);
        }
        sync;
        double *temp=y; y=z; z=temp;
    }
    if(yorig!=y){
        double *temp=y; y=z; z=temp;
        memcpy(z,y,N*sizeof(double));
    }
    for(int c=ncpu-1;c>=0;c--){
        free(wb[c][1]-2*ncpu); free(wb[c][0]-2*ncpu);
    }
}

double lorenz(){
    if(K%2){
        fprintf(stderr,"K=%d but must be even!\n",K);
        exit(1);
    }
    xgoldb=mymalloc((N+3)*sizeof(double),"xgoldb");
    xb=mymalloc((N+3*nprocs)*sizeof(double),"xb");
    yb=mymalloc((N+3*nprocs)*sizeof(double),"yb");
    double *x0=mymalloc(N*sizeof(double),"x0");
    x=xb+2*nprocs; xgold=xgoldb+2; y=yb+2*nprocs;
    double ops=6.0*K*N;
    dt=(double)(T)/K;
    expmdt=exp(-dt);
    forcdt=2*exp(-dt/2)*sinh(dt/2)*F;
    for(int i=0;i<N;i++) x0[i]=sin(2*M_PI*i/N);
    memcpy(xgold,x0,N*sizeof(double));
    lorenzserial(xgold);
    double tmin=1e32;
    int nmin=nprocs;
    for(ncpu=nprocs;ncpu>=1;ncpu=ncpu*FACTOR){
        workers(ncpu) {
            double ttot=0.0;
            for(int j=0;;j++){
                memcpy(x,x0,N*sizeof(double));
                tic();
                if(ncpu==1) lorenzserial(x);
                else lorenzparallel(x);
                double t=toc();
                if(t<0) {
                    printf("oops time went backwards by %g!\n",t);
                    exit(1);
                }
                for(int i=0;i<N;i++) if(x[i]!=xgold[i]) {
                    fprintf(stderr,
                    "Error %g in parallel solver at index %d!\n",
                    fabs(x[i]-xgold[i]),i);
                    exit(1);
                }
                if(t<tmin){ tmin=t; nmin=ncpu; }
                ttot+=t;
                if(ttot>5 && j>3) break;
            }
        }
    }
    printf("N=%d K=%d Workers=%d Sec=%g Mflops=%g",
        N,K,nmin,tmin,ops/tmin/1e6);
    if(fstress){
        printf("\n%21sPerforming %g minute stress test...","",fstress);
        fflush(stdout);
        ncpu=nmin;
        workers(ncpu) {
            tic();
            while(toc()<fstress*60){
                memcpy(x,x0,N*sizeof(double));
                if(ncpu==1) lorenzserial(x);
                else lorenzparallel(x);
            }
        }
        printf("done");
    }
    free(x0); free(yb); free(xgoldb); free(xb);
    return ops/tmin/1e6;
}
